"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const lodash_1 = require("lodash");
const request_promise_1 = __importDefault(require("request-promise"));
const types_1 = require("./types");
const https_1 = require("https");
const helpers_1 = require("../helpers");
const url_1 = require("url");
const querystring_1 = __importDefault(require("querystring"));
// eslint-disable-next-line @typescript-eslint/no-var-requires
const CFError = require('cf-errors');
const logger = helpers_1.createNewLogger('codefresh:infra:git-providers:bitbucket');
const LIMIT_PER_PAGE = 100;
const MAX_PAGE = 200;
const MAX_PAGES = 30;
const _cleanEncodedFilePath = (filepath) => {
    const prefix = './';
    const hasPrefix = filepath.indexOf(prefix) === 0;
    const path = hasPrefix ? filepath.slice(prefix.length) : filepath;
    return encodeURIComponent(path);
};
const _buildError = (response) => {
    let errStr = lodash_1.get(response, 'body.error.message');
    if (lodash_1.get(response, 'body.error.id')) {
        errStr = `${errStr} [${lodash_1.get(response, 'body.error.id')}]`;
    }
    return new types_1.HttpError(response.statusCode, errStr);
};
const _toRepo = (rawRepo, issues) => {
    const isGroup = lodash_1.get(rawRepo, 'owner.type') === 'team';
    const httpsClone = rawRepo.links.clone.find((link) => link.name === 'https');
    const sshClone = rawRepo.links.clone.find((link) => link.name === 'ssh');
    return {
        id: rawRepo.uuid,
        provider: 'bitbucket',
        name: rawRepo.name,
        full_name: rawRepo.full_name,
        private: rawRepo.is_private,
        pushed_at: rawRepo.updated_on,
        open_issues: issues ? issues : 0,
        clone_url: lodash_1.get(httpsClone, 'href', ''),
        ssh_url: lodash_1.get(sshClone, 'href', ''),
        owner: {
            login: lodash_1.get(rawRepo, 'owner.display_name', ''),
            avatar_url: lodash_1.get(rawRepo, 'owner.links.avatar.href', ''),
            creator: null,
        },
        org: isGroup ? lodash_1.get(rawRepo, 'owner.username') : null,
        default_branch: lodash_1.get(rawRepo, 'mainbranch.name', ''),
        permissions: {
            admin: true,
        },
        webUrl: lodash_1.get(rawRepo, 'links.html.href', ''),
    };
};
const _toBranch = (rawBranch) => {
    return {
        name: rawBranch.name,
        id: rawBranch.name,
        commit: {
            sha: rawBranch.target.hash,
            commiter_name: lodash_1.get(rawBranch, 'target.author.user.display_name', ''),
            message: lodash_1.get(rawBranch, 'target.message', ''),
            url: lodash_1.get(rawBranch, 'target.links.html.href', ''),
        }
    };
};
const _toUser = (rawUser, rawEmails) => {
    var _a, _b, _c, _d;
    let isLastPrimary = false;
    let lastEmailInfo;
    for (let i = 0; i < rawEmails.length; i++) {
        const emailInfo = rawEmails[i];
        isLastPrimary = lastEmailInfo === null || lastEmailInfo === void 0 ? void 0 : lastEmailInfo.is_primary;
        if (emailInfo.is_confirmed && emailInfo.is_primary) {
            lastEmailInfo = emailInfo;
            break;
        }
        if (emailInfo.is_primary) {
            lastEmailInfo = emailInfo;
        }
        if ((lastEmailInfo && !isLastPrimary || !lastEmailInfo) && emailInfo.is_confirmed) {
            lastEmailInfo = emailInfo;
        }
    }
    return {
        login: rawUser.username,
        avatar_url: (_b = (_a = rawUser === null || rawUser === void 0 ? void 0 : rawUser.links) === null || _a === void 0 ? void 0 : _a.avatar) === null || _b === void 0 ? void 0 : _b.href,
        email: (lastEmailInfo === null || lastEmailInfo === void 0 ? void 0 : lastEmailInfo.email) || '',
        web_url: ((_d = (_c = lastEmailInfo === null || lastEmailInfo === void 0 ? void 0 : lastEmailInfo.links) === null || _c === void 0 ? void 0 : _c.self) === null || _d === void 0 ? void 0 : _d.href) || ''
    };
};
class Bitbucket {
    constructor(opts) {
        this.apiPrefix = 'api/2.0/';
        this.baseUrl = opts.apiURL || 'https://bitbucket.org/';
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`;
        }
        if (this.baseUrl === 'https://api.bitbucket.org/') {
            this.apiPrefix = '2.0/';
        }
        else if (this.baseUrl.endsWith('2.0/')) {
            // already has apiPrefix in url
            this.apiPrefix = '';
        }
        this.timeout = opts.timeout || 10000;
        if (opts.username && opts.username !== 'x-token-auth') {
            // eslint-disable-next-line @typescript-eslint/naming-convention
            this.authenticationHeader = { Authorization: `Basic ${Buffer.from(`${opts.username}:${opts.password}`).toString('base64')}` };
        }
        else {
            // eslint-disable-next-line @typescript-eslint/naming-convention
            this.authenticationHeader = { Authorization: `Bearer ${opts.password}` };
        }
        if (this.baseUrl.startsWith('https') && (process.env.NODE_TLS_REJECT_UNAUTHORIZED === '0' || opts.insecure)) {
            this.agent = new https_1.Agent({ rejectUnauthorized: false });
        }
    }
    getName() {
        return 'bitbucket';
    }
    async performAPICall(opts) {
        const method = opts.method || 'GET';
        const requestHeaders = lodash_1.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {});
        const requestOptions = {
            method,
            url: `${this.baseUrl}${this.apiPrefix}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            form: opts.data,
            resolveWithFullResponse: true,
            agent: this.agent,
            simple: false,
        };
        return request_promise_1.default(requestOptions);
    }
    async paginate(opts, predicate) {
        let start = opts.page || 1;
        let isLastPage = false;
        const results = [];
        const limit = opts.limit || Number.MAX_SAFE_INTEGER;
        while (!isLastPage && start <= MAX_PAGES && results.length < limit) {
            opts.qs = lodash_1.merge(opts.qs, {
                page: '' + start,
                pagelen: '' + (limit - results.length > LIMIT_PER_PAGE ? LIMIT_PER_PAGE : limit - results.length),
            });
            const res = await this.performAPICall(opts);
            if (res.statusCode >= 400) {
                return res;
            }
            if (predicate) {
                const predicateResult = lodash_1.find(res.body.values, predicate);
                if (predicateResult) {
                    // found the wanted result
                    return {
                        statusCode: 200,
                        body: predicateResult,
                    };
                }
            }
            res.body.values.forEach((result) => { results.push(result); });
            isLastPage = !lodash_1.has(res.body, 'next');
            start += 1;
        }
        return {
            statusCode: 200,
            body: results,
        };
    }
    async paginateWithNext(opts) {
        var _a;
        let isLastPage = false;
        const results = [];
        const limit = opts.limit || Number.MAX_SAFE_INTEGER;
        opts.qs = lodash_1.merge(opts.qs, { pagelen: limit });
        const nextOpts = { ...opts };
        for (let i = 1; i <= MAX_PAGE && !isLastPage; i++) {
            const res = await this.performAPICall(nextOpts);
            const nextUrl = (_a = res.body) === null || _a === void 0 ? void 0 : _a.next;
            isLastPage = !nextUrl;
            if (nextUrl) {
                const url = new url_1.URL(nextUrl);
                nextOpts.qs = querystring_1.default.parse(url.searchParams.toString());
                url.search = '';
                nextOpts.url = url.toString();
            }
            if (res.statusCode >= 400) {
                return res;
            }
            if (opts.page && i < opts.page) {
                // skip all pages until we reach our desired page
                continue;
            }
            const resultsLeft = limit - results.length;
            res.body.values.slice(0, resultsLeft).forEach((result) => results.push(result));
            if (results.length >= limit) {
                break;
            }
        }
        return {
            statusCode: 200,
            body: results,
        };
    }
    async createRepository(opt) {
        const repoRes = await this.performAPICall({
            api: `repositories/${opt.owner}/${opt.repo}`,
            method: 'POST',
            json: true,
            data: {
                is_private: opt.private,
            }
        });
        if (repoRes.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(repoRes),
                message: `Failed to create repository: ${JSON.stringify(opt)}`
            });
        }
        return _toRepo(repoRes.body, 0);
    }
    async fetchRawFile(opts) {
        const res = await this.performAPICall({
            api: `repositories/${opts.owner}/${opts.repo}/src/${opts.ref}/${_cleanEncodedFilePath(opts.path)}`,
        });
        if (res.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(res),
                message: `Failed to retrieve file: ${JSON.stringify(opts)}`,
            });
        }
        return res.body;
    }
    async getBranch(opts) {
        const res = await this.performAPICall({
            api: `repositories/${opts.owner}/${opts.repo}/refs/branches/${opts.branch}`,
            json: true
        });
        if (res.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(res),
                message: `Failed to retrieve branch: ${JSON.stringify(opts)}`,
            });
        }
        return _toBranch(res.body);
    }
    async getRepository(opts) {
        const repoPromise = this.performAPICall({
            api: `repositories/${opts.owner}/${opts.repo}`,
            json: true,
        });
        const issuesPromise = this.performAPICall({
            api: `repositories/${opts.owner}/${opts.repo}/issues`,
            json: true,
        });
        const [repoResponse, issuesResponse] = await Promise.all([repoPromise, issuesPromise]);
        if (repoResponse.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(repoResponse),
                message: `Failed to get repository: ${JSON.stringify(opts)}`,
            });
        }
        if (issuesResponse.statusCode >= 400) {
            logger.warn(`Failed to get repository issues: ${JSON.stringify(opts)}, ${_buildError(issuesResponse)}`);
            issuesResponse.body.size = 0; // have 0 issues
        }
        return _toRepo(repoResponse.body, issuesResponse.body.size);
    }
    async listBranches(opts) {
        const res = await this.paginate({
            api: `repositories/${opts.owner}/${opts.repo}/refs/branches`,
            json: true,
            page: opts.page,
            limit: opts.limit,
        });
        if (res.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(res),
                message: `Failed to get branches list: ${JSON.stringify(opts)}`,
            });
        }
        return res.body.map(_toBranch);
    }
    async listRepositoriesForOwner(opt) {
        throw new Error('Method not implemented.');
    }
    async createRepositoryWebhook(opt) {
        throw new Error('Method not implemented.');
    }
    async listWebhooks(opt) {
        throw new Error('Method not implemented.');
    }
    async deleteRepositoryWebhook(opt) {
        throw new Error('Method not implemented.');
    }
    async listRepositoriesWithAffiliation(opt) {
        const { limit = 25, page = 1, /*direction = 'desc',*/ affiliation = 'member' } = opt;
        const res = await this.paginateWithNext({
            api: `repositories`,
            page: page,
            limit: limit,
            json: true,
            qs: {
                // Can't sort because it is broken atm for Bitbucket
                // sort: (direction === 'desc' ? '-' : '') + 'updated_on',
                role: affiliation,
            }
        });
        if (res.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(res),
                message: `Failed to get repositories list: ${JSON.stringify(opt)}`,
            });
        }
        return res.body.map(_toRepo);
    }
    async listOrganizations(opt) {
        const { limit = 25, page = 1 } = opt;
        const res = await this.paginate({
            api: 'user/permissions/workspaces',
            page: page,
            limit: limit,
            json: true,
        });
        if (res.statusCode >= 400) {
            throw new CFError({
                cause: _buildError(res),
                message: `Failed to get organization list: ${JSON.stringify(opt)}`,
            });
        }
        return res.body
            .filter((i) => i.permission !== 'member') // filter workspaces where you can't create repos
            .map((i) => i.workspace.slug);
    }
    async listRepositoriesForOrganization(opt) {
        throw new Error('Method not implemented.');
    }
    async createCommitStatus(opt) {
        throw new Error('Method not implemented.');
    }
    async getUser(opt) {
        if (!(opt === null || opt === void 0 ? void 0 : opt.username)) {
            const userPromise = this.performAPICall({
                api: `user`,
                json: true,
            });
            const emailsPromise = this.performAPICall({
                api: `user/emails`,
                json: true,
            });
            const [userRes, emailsRes] = await Promise.all([userPromise, emailsPromise]);
            return _toUser(userRes.body, emailsRes.body.values);
        }
        throw new Error('bitbucket cloud doesnt support get user by username');
    }
    async getUserByEmail(email) {
        throw new Error('Method not implemented.');
    }
    async getPullRequestFiles(opt) {
        throw new Error('Method not implemented.');
    }
    async getPullRequest(opt) {
        throw new Error('Method not implemented.');
    }
    async getRepositoryPermissions(opt) {
        return {
            read: true,
            write: true,
        };
    }
    async assertApiScopes(opt) {
        //TODO : Impl validation
    }
    toOwnerRepo(fullRepoName) {
        const [owner, ...repoParts] = fullRepoName.split('/');
        return [owner, repoParts.join('/')];
    }
    getAuth() {
        return {
            headers: this.authenticationHeader,
        };
    }
    isTokenMutable() {
        return false;
    }
}
exports.default = Bitbucket;
//# sourceMappingURL=bitbucket.js.map